(defconst cobol--imenu-datadiv-sections
  (cobol--with-opt-whitespace-line
   "\\(file\\|working-storage\\|local-storage\\|linkage\\|report\\|screen\\)"
   "[ \t]+section[ \t]*\\.")
  "Regexp which matches a section in the data division.")

(defconst cobol--imenu-procdiv-ignore-re
  (cobol--with-opt-whitespace-line
   (regexp-opt (append cobol-scope-terminators '("exit" "goback"))))
  ;; (cobol--with-opt-whitespace-line "\\(end-.*\\|exit\\.\\|goback \\)")
  "Things that match `cobol--procedure-re' which we don't want to add to the
  index.")


(defconst cobol--imenu-identification-division-re
  "^[ \t]*identification[ \t]+division"
  "Regexp to match the identification division")

(defconst cobol--imenu-program-name
  "\\(function\\|program\\)-id\\.[ \t]+\\(\\w+\\)[ \t]*\\."
  "Regexp to find the name of a COBOL (sub)program.")

(defun cobol--imenu-section-abbreviation (section)
  "Returns an abbreviation of the section name SECTION.

This is to not make the menu entries too wide when there are subprograms whose
  name is prepended to the section."
  (let ((section-down (downcase section)))
    (cond ((string= section-down "file")
           " FS")
          ((string= section-down "working-storage")
           " WSS")
          ((string= section-down "local-storage")
           " LSS")
          ((string= section-down "linkage")
           " LS")
          ((string= section-down "report")
           " RS")
          ((string= section-down "screen")
           " SS"))))

(defun cobol--imenu-create-datadiv-index ()
  "Create an Imenu sublist item for a section in the data division.

Leaves point at the start of the next section heading in the data
  division, or at the procedure division heading if there were no
  more. Or eob if there is neither."
  (let ((idx-elms nil))
    ;; We stop when we get to the next datadiv section, or when we get to the
    ;; procedure division, or when we hit eob
    (while (and (not (looking-at-p cobol--imenu-datadiv-sections))
                (not (looking-at-p cobol--procedure-division-re))
                (not (eobp)))
      (forward-line 1)
      ;; When we're looking at a data item declaration
      (when (looking-at
             "^[ \t]*\\([[:digit:]]\\{1,2\\}\\|[fsr]d\\)[ \t]+\\(\\w+\\)")
        ;; Record the level number and name
        (let ((level (match-string-no-properties 1))
              (name (match-string-no-properties 2)))
          ;; If it is not just a filler, put it in the list
          (when (not (string= (downcase name) "filler"))
            (push (cons (concat level " " name) (match-beginning 2))
                  idx-elms)))))
    ;; Reverse the list so it is in the same order as in the buffer
    (nreverse idx-elms)))


(defun cobol--imenu-create-procdiv-index ()
  "Create imenu index for the procedure division."
  ;; Search for the procedure division start
  (re-search-forward cobol--procedure-division-re nil t)
  (let ((idx-elms nil))
    ;; While we're not at the end of buffer or at the start of the next
    ;; subprogram or function
    (while (and (not (eobp))
                (not (looking-at-p cobol--imenu-identification-division-re)))
      (forward-line 1)
      (when (and (looking-at cobol--procedure-re)
                 (not (looking-at-p cobol--imenu-procdiv-ignore-re)))
        (push (cons (match-string-no-properties 1) (match-beginning 1))
              idx-elms)))
    (nreverse idx-elms)))


(defun cobol--imenu-create-index ()
  "Create an Imenu index for the current buffer.

The first program in the buffer gets top level entries for each section in the
  data division and one for the procedure division.  If there are more
  subprograms or functions in the buffer, they will have their entries prefixed
  by the subprogram/function name and there will be a submenu listing all the
  subprograms/functions."
  (goto-char (point-min))
  (let ((case-fold-search t)
        (index nil)                     ;The imenu index
        (subprograms nil))              ;List of subprograms
    ;; While there are more subprograms/functions
    (while (re-search-forward cobol--imenu-identification-division-re nil t)
      (re-search-forward cobol--imenu-program-name nil t)
      (let ((subprogram-name (match-string-no-properties 2))
            (subprogram-pos (match-beginning 1)))
        ;; While we're not at the start of the procedure division, search
        ;; forward for a section in the data division
        (while (and (not (looking-at-p cobol--procedure-division-re))
                    (re-search-forward cobol--imenu-datadiv-sections nil t))
          (let ((section (match-string-no-properties 1))
                (subidx (cobol--imenu-create-datadiv-index)))
            (when subidx
              (push
               (cons
                ;; If there are multiple subprograms (ie this is not the first
                ;; time we get here) make the entry the suprogram name followed
                ;; by an abbreviation of the section name. Otherwise (ie for the
                ;; first program in the buffer) make the entry the full section
                ;; name
                (if subprograms
                    (concat subprogram-name 
                            (cobol--imenu-section-abbreviation section))
                  (concat (capitalize section) " Section"))
                subidx)
               index))))
        (let ((procdiv-idx (cobol--imenu-create-procdiv-index)))
          (when procdiv-idx
            (push (cons (if subprograms (concat subprogram-name " PD")
                          "Procedure Division")
                        procdiv-idx)
                  index)))
        (push (cons subprogram-name subprogram-pos) subprograms)))
    ;; If there were multiple programs defined in this file make a submenu with
    ;; their names
    (when (> (length subprograms) 1)
      (push (cons "Subprograms" (nreverse subprograms)) index))
    (nreverse index)))

